@@ -1,22 +1,16 @@
 from itertools import combinations
 import random
+
+import pandas
 import pingouin as pg
 from copy import deepcopy
 
 import numpy as np
 from graphviz import Digraph
 from numba import njit
-from rpy2.robjects.packages import importr
-from rpy2.robjects import pandas2ri
-import rpy2.rinterface
 from scipy.stats.distributions import chi2
-import scipy.stats as st
 from .score import compute_weights
 
-base, bnlearn, wtd = importr('base'), importr('bnlearn'), importr('weights')
-
-pandas2ri.activate()
-
 
 # random.seed(941214)
 
@@ -30,60 +24,61 @@ def miss_mechanism(dag, noise='MAR', rom=0.5):
     :param rom: ratio of missing variables
     :return: missing mechanism
     '''
+    if type(dag) is str:
+        dag = from_bnlearn(dag)
+    elif type(dag) is not dict:
+        raise Exception('The format of input DAG is invalid.')
+    varnames = list(dag.keys())
     cause_dict = {}
     if noise == 'MCAR':
-        vars_miss = random.sample(list(bnlearn.nodes(dag)), round(len(bnlearn.nodes(dag)) * rom))
+        vars_miss = random.sample(varnames, round(len(varnames) * rom))
         for var in vars_miss:
             cause_dict[var] = []
     elif noise == 'MAR':
-        varnames = list(bnlearn.nodes(dag))
         nom = round(len(varnames) * rom)
         noc = len(varnames) - nom
-        vstructs = np.array(bnlearn.vstructs(dag))
-        vstructs = vstructs.reshape(3, int(len(vstructs) / 3))
+        vstructs = vstructures(dag)
         vars_miss = []
         vars_comp = []
-        for i in range(vstructs.shape[1]):
-            if len(vars_comp) != noc or (len(vars_comp) == noc and vstructs[1, i] in vars_comp):
-                if vstructs[0, i] not in vars_miss + vars_comp:
-                    cause_dict[vstructs[0, i]] = [vstructs[1, i]]
-                    vars_miss.append(vstructs[0, i])
-                    vars_comp = list(set(vars_comp + [vstructs[1, i]]))
-                elif vstructs[2, i] not in vars_miss + vars_comp:
-                    cause_dict[vstructs[2, i]] = [vstructs[1, i]]
-                    vars_miss.append(vstructs[2, i])
-                    vars_comp = list(set(vars_comp + [vstructs[1, i]]))
+        for i in range(len(vstructs)):
+            if len(vars_comp) != noc or (len(vars_comp) == noc and vstructs[i][1] in vars_comp):
+                if vstructs[i][0] not in vars_miss + vars_comp:
+                    cause_dict[vstructs[i][0]] = [vstructs[i][1]]
+                    vars_miss.append(vstructs[i][0])
+                    vars_comp = list(set(vars_comp + [vstructs[i][1]]))
+                elif vstructs[i][2] not in vars_miss + vars_comp:
+                    cause_dict[vstructs[i][2]] = [vstructs[i][1]]
+                    vars_miss.append(vstructs[i][2])
+                    vars_comp = list(set(vars_comp + [vstructs[i][1]]))
                 if len(vars_miss) == nom:
                     break
         if len(vars_miss) < nom:
             vars_miss2 = random.sample([x for x in varnames if x not in vars_miss + vars_comp], nom - len(vars_miss))
             vars_comp = [x for x in varnames if x not in vars_miss + vars_miss2]
             for var in vars_miss2:
-                nbr = list(bnlearn.nbr(dag, var))
+                nbr = adjacent(dag, var)
                 if any(item in vars_comp for item in nbr):
                     cause_dict[var] = random.sample([x for x in nbr if x in vars_comp], 1)
                 else:
                     cause_dict[var] = random.sample(vars_comp, 1)
     elif noise == 'MNAR':
-        varnames = list(bnlearn.nodes(dag))
         nom = round(len(varnames) * rom)
-        vstructs = np.array(bnlearn.vstructs(dag))
-        vstructs = vstructs.reshape(3, int(len(vstructs) / 3))
+        vstructs = vstructures(dag)
         vars_miss = []
-        for i in range(vstructs.shape[1]):
-            if vstructs[0, i] not in cause_dict:
-                cause_dict[vstructs[0, i]] = [vstructs[1, i]]
-                vars_miss = list(set(vars_miss + [vstructs[0, i], vstructs[1, i]]))
-            elif vstructs[2, i] not in cause_dict:
-                cause_dict[vstructs[2, i]] = [vstructs[1, i]]
-                vars_miss = list(set(vars_miss + [vstructs[2, i], vstructs[1, i]]))
+        for i in range(len(vstructs)):
+            if vstructs[i][0] not in cause_dict:
+                cause_dict[vstructs[i][0]] = [vstructs[i][1]]
+                vars_miss = list(set(vars_miss + [vstructs[i][0], vstructs[i][1]]))
+            elif vstructs[i][2] not in cause_dict:
+                cause_dict[vstructs[i][2]] = [vstructs[i][1]]
+                vars_miss = list(set(vars_miss + [vstructs[i][2], vstructs[i][1]]))
             if len(vars_miss) >= nom:
                 break
         if len(vars_miss) < nom:
             vars_miss2 = random.sample([x for x in varnames if x not in vars_miss], nom - len(vars_miss))
             vars_comp = [x for x in varnames if x not in vars_miss + vars_miss2]
             for var in vars_miss2:
-                nbr = list(bnlearn.nbr(dag, var))
+                nbr = adjacent(dag, var)
                 if any(item not in vars_comp for item in nbr):
                     cause_dict[var] = random.sample([x for x in nbr if x not in vars_comp + [var]], 1)
                 else:
@@ -93,6 +88,22 @@ def miss_mechanism(dag, noise='MAR', rom=0.5):
     return cause_dict
 
 
+def vstructures(dag):
+    if type(dag) is str:
+        dag = from_bnlearn(dag)
+    elif type(dag) is not dict:
+        raise Exception('The format of input dag is invalid')
+
+    cpdag = dag2cpdag(dag)
+    vstructs = []
+    for var in cpdag:
+        if len(cpdag[var]['par']) > 1:
+            for pair in [(a, b) for idx, a in enumerate(cpdag[var]['par']) for b in cpdag[var]['par'][idx + 1:]]:
+                if pair[0] not in adjacent(cpdag, pair[1]):
+                    vstructs.append([pair[0], var, pair[1]])
+    return vstructs
+
+
 # add missing value in dataset
 def add_missing(data, cause_dict):
     data_missing = deepcopy(data)
@@ -235,17 +246,18 @@ def to_bnlearn(dag):
                 output += par + ':'
             output = output[:-1]
         output += ']'
-    return bnlearn.model2network(output)
+    return output
 
 
 # convert bnlearn format to my format
 def from_bnlearn(dag):
+    dag = dag[1: len(dag) - 1]
     output = {}
-    for node in bnlearn.nodes(dag):
-        output[node] = {}
-        output[node]['par'] = list(bnlearn.parents(dag, node))
-        output[node]['nei'] = list(
-            base.setdiff(base.setdiff(bnlearn.nbr(dag, node), bnlearn.parents(dag, node)), bnlearn.children(dag, node)))
+    for node in dag.split(']['):
+        if '|' not in node:
+            output[node] = {'par': [], 'nei': []}
+        else:
+            output[node.split('|')[0]] = {'par': node.split('|')[1].split(':'), 'nei': []}
     return output
 
 
@@ -469,38 +481,24 @@ def g_counter(data, arities, cols, weights=None):
 
 
 # statistical fisher's z test
-def zf_test(data, cols, weights=None):
+def z_test(data, varnames, cols):
+    data = pandas.DataFrame(data, columns=varnames)
     if len(cols) == 2:
-        rho = wtd.partial_cor(x=data.columns[cols[0]], y=data.columns[cols[1]], weight=weights) if weights == None else \
-            pg.partial_corr(data, data.columns[cols[0]], data.columns[cols[1]])['r'][0]
+        return pg.partial_corr(data, varnames[cols[0]], varnames[cols[1]])['p-val'][0]
     elif len(cols) > 2:
-        rho = wtd.partial_cor(x=data.columns[cols[0]], y=data.columns[cols[1]], preds=list(data.columns[cols[2:]]),
-                              weight=weights) if weights == None else \
-            pg.partial_corr(data, data.columns[cols[0]], data.columns[cols[1]], list(data.columns[cols[2:]]))['r'][0]
-    else:
-        raise Exception('Length of input cols is less than 2')
-    z = np.arctanh(rho) * np.sqrt(data.shape[0] - 3 - len(cols) + 2)
-    return st.norm.sf(abs(z)) * 2
-
-
-# statistical pearson test
-def pearson_test(data, cols):
-    if len(cols) == 2:
-        return pg.partial_corr(data, data.columns[cols[0]], data.columns[cols[1]])['p-val'][0]
-    elif len(cols) > 2:
-        return \
-            pg.partial_corr(data, data.columns[cols[0]], data.columns[cols[1]], list(data.columns[cols[2:]]))['p-val'][
-                0]
+        return pg.partial_corr(data, varnames[cols[0]], varnames[cols[1]], [varnames[i] for i in cols[2:]])['p-val'][0]
     else:
         raise Exception('Length of input cols is less than 2')
 
 
 # plot the bnlearn DAG
 def plot(dag, filename):
+    if type(dag) is not dict:
+        dag = from_bnlearn(dag)
     dot = Digraph()
-    for node in bnlearn.nodes(dag):
+    for node in dag:
         dot.node(node)
-        for parent in bnlearn.parents(dag, node):
+        for parent in dag[node]['par']:
             if parent not in dot.body:
                 dot.node(parent)
             dot.edge(parent, node)
@@ -529,13 +527,6 @@ def pdag2dag(pdag):
     return dag
 
 
-def pdag2cpdag(pdag):
-    dag = pdag2dag(pdag)
-    dag_bnlearn = to_bnlearn(dag)
-    cpdag_bnlearn = bnlearn.cpdag(dag_bnlearn)
-    return from_bnlearn(cpdag_bnlearn)
-
-
 # check function used for pdag2dag
 def pdag_check(pdag, x):
     for y in pdag[x]['nei']:
@@ -547,18 +538,10 @@ def pdag_check(pdag, x):
     return True
 
 
-# convert gobnilp cpdag to bnlearn format
-def gobnilp2bnlearn(dag):
-    try:
-        return bnlearn.cpdag(bnlearn.model2network(dag))
-    except rpy2.rinterface.embedded.RRuntimeError:
-        return None
-
-
 # convert DAG into CPDAG
 def dag2cpdag(dag):
     # initialise CPDAG
-    cpdag = {var: {'par': [], 'nei': []} for var in dag}
+    cpdag = {var: {'par': [], 'nei': [v for v in dag[var]['nei']]} for var in dag}
     # remain all unshielded colliders
     for var in dag:
         if len(dag[var]['par']) > 1:
@@ -648,4 +631,123 @@ def dag2cpdag(dag):
             if not stop_flag:
                 break
         if stop_flag:
-            return cpdag
\ No newline at end of file
+            return cpdag
+
+
+# statistical test
+def independence_test(data, cols, varnames=None, weights=None, test='g-test', threshold=0.05):
+    '''
+    :param data: the unique datapoints as a 2-d array, each row is a datapoint, assumed unique
+    :param cols: the columns (=variables) for the marginal contingency table. columns must be ordered low to high
+    :param test: the name of test
+    :param threshold: the threshold for G-test
+
+    :returns 1st element: test result
+    :returns 2nd element: test score
+    '''
+    if test == 'g-test':
+        p = g_test(data, cols, weights)
+    elif test == 'z-test':
+        p = z_test(data, varnames, cols)
+    else:
+        raise Exception('Undefined test function.')
+    return p > threshold, p
+
+
+# orient v-structrues
+def orient_v_structure(data, varnames, pc, sepset, test, threshold=0.05):
+    dag = {}
+    v_candidate = {}
+    for var in pc:
+        dag[var] = {}
+        dag[var]['par'] = []
+        dag[var]['nei'] = [v for v in pc[var]]
+        if len(pc[var]) > 1:
+            pc_copy = [v for v in pc[var]]
+            for var1 in pc_copy:
+                for var2 in pc_copy[pc_copy.index(var1) + 1:]:
+                    if var2 in sepset[var1]:
+                        if var not in sepset[var1][var2]:
+                            cols = np.array([varnames.index(v) for v in [var1, var2] + list(sepset[var1][var2])])
+                            v_candidate[tuple([var, var1, var2])] = \
+                                independence_test(data, cols, varnames=varnames, test=test, threshold=threshold)[1]
+    while len(v_candidate):
+        pair = min(v_candidate, key=v_candidate.get)
+        if (pair[0] not in dag[pair[1]]['par']) & (pair[0] not in dag[pair[2]]['par']):
+            if (sin_path_check(dag, pair[0], pair[1])[1] != [pair[0], pair[1]]) & \
+                    (sin_path_check(dag, pair[0], pair[2])[1] != [pair[0], pair[2]]):
+                if pair[1] in dag[pair[0]]['nei']:
+                    dag[pair[0]]['par'].append(pair[1])
+                    dag[pair[0]]['nei'].remove(pair[1])
+                    dag[pair[1]]['nei'].remove(pair[0])
+                if pair[2] in dag[pair[0]]['nei']:
+                    dag[pair[0]]['par'].append(pair[2])
+                    dag[pair[0]]['nei'].remove(pair[2])
+                    dag[pair[2]]['nei'].remove(pair[0])
+        del v_candidate[pair]
+    return dag
+
+
+# orientation propagation
+def orient_propagation(dag):
+    while True:
+        dag_check = {var: {'par': [v for v in dag[var]['par']], 'nei': [v for v in dag[var]['nei']]} for var in dag}
+        # check cycles
+        for var in dag:
+            if len(dag_check[var]['nei']):
+                nei_set = [v for v in dag_check[var]['nei']]
+                for nei in nei_set:
+                    sin_flag, sin_direction = sin_path_check(dag, var, nei)
+                    v_flag, v_direction = v_check(dag_check, var, nei)
+                    if sin_flag:
+                        if (not v_flag) | (sin_direction == v_direction):
+                            if sin_direction[0] not in dag[sin_direction[1]]['par']:
+                                dag[sin_direction[1]]['par'].append(sin_direction[0])
+                            if sin_direction[0] in dag[sin_direction[1]]['nei']:
+                                dag[sin_direction[1]]['nei'].remove(sin_direction[0])
+                                dag[sin_direction[0]]['nei'].remove(sin_direction[1])
+        # check v-structures
+        for var in dag:
+            if len(dag_check[var]['nei']):
+                nei_set = [v for v in dag_check[var]['nei']]
+                for nei in nei_set:
+                    sin_flag, sin_direction = sin_path_check(dag, var, nei)
+                    v_flag, v_direction = v_check(dag_check, var, nei)
+                    if v_flag & (v_direction != 'both'):
+                        if (not sin_flag) | (sin_direction == v_direction):
+                            if v_direction[0] not in dag[v_direction[1]]['par']:
+                                dag[v_direction[1]]['par'].append(v_direction[0])
+                            if v_direction[0] in dag[v_direction[1]]['nei']:
+                                dag[v_direction[1]]['nei'].remove(v_direction[0])
+                                dag[v_direction[0]]['nei'].remove(v_direction[1])
+
+        # check Meek's rule 3
+        for var in dag:
+            if len(dag_check[var]['nei']):
+                nei_set = [v for v in dag_check[var]['nei']]
+                for nei in nei_set:
+                    if (not sin_path_check(dag_check, var, nei)[0]) & (not v_check(dag_check, var, nei)[0]):
+                        common_var = list(set(dag[nei]['par']).intersection(dag[var]['nei']))
+                        if len(common_var) > 1:
+                            flag = False
+                            for var1 in common_var:
+                                common_var_remain = deepcopy(common_var)
+                                common_var_remain.remove(var1)
+                                for var2 in common_var_remain:
+                                    if (var1 not in (dag[var2]['par'] + dag[var2]['nei'])) & (
+                                            var2 not in dag[var1]['par']):
+                                        flag = True
+                                        dag[var]['nei'].remove(nei)
+                                        dag[nei]['nei'].remove(var)
+                                        dag[nei]['par'].append(var)
+                                        break
+                                if flag:
+                                    break
+        if dag_check == dag:
+            break
+    return dag
+
+
+# get the adjacent variables of a certain node in a graph
+def adjacent(graph, var):
+    return graph[var]['par'] + graph[var]['nei'] + [v for v in graph if var in graph[v]['par']]
